use std::{
    fs::File,
    io::{Read, Write},
};

use crate::{
    dynamic_config::{rehash_config, DynamicConfigLock},
    AcmeConfig, ServerConfig, TlsConfig,
};
use actix_web::{
    error::ErrorNotFound,
    get,
    web::{self, Data},
    Responder,
};
use anyhow::bail;
use instant_acme::{
    Account, AuthorizationStatus, ChallengeType, Identifier, NewAccount, NewOrder, OrderStatus,
};
use log::{error, info, warn};
use rcgen::{CertificateParams, DistinguishedName, KeyPair};
use time::{Duration, OffsetDateTime};
use tokio::time::sleep;
use x509_parser::pem::parse_x509_pem;

// This explicitly checks not_after, since it is for renewal - we don't want
// to renew because a certificate isn't valid yet etc...
fn get_current_expiry(certfile: &str) -> anyhow::Result<OffsetDateTime> {
    let mut contents = Vec::new();
    File::open(certfile)?.read_to_end(&mut contents)?;
    let pem = parse_x509_pem(&contents)?.1;
    Ok(pem.parse_x509()?.validity.not_after.to_datetime())
}

async fn try_renew_certificate(
    config_file: &str,
    dynconfig: &DynamicConfigLock,
    tls: &TlsConfig,
    acme: &AcmeConfig,
) -> anyhow::Result<()> {
    let mut contents = String::new();
    let account: Account = match File::open(&acme.acme_account_file)
        .and_then(|mut f| f.read_to_string(&mut contents))
        .map_err(anyhow::Error::from)
        .and_then(|_| serde_json::from_str(&contents).map_err(anyhow::Error::from))
    {
        Ok(creds) => Account::from_credentials(creds).await?,
        Err(e) => {
            info!(
                "Can't load existing credentials ({}), going to create new ones.",
                e
            );
            if !acme.contact_email.contains('@') {
                bail!("contact_email configuration not set to a valid email");
            }
            let mut email_url = acme.contact_email.to_owned();
            if !email_url.starts_with("mailto:") {
                email_url = format!("mailto:{}", email_url);
            }
            let (account, account_creds) = instant_acme::Account::create(
                &NewAccount {
                    contact: &[&email_url],
                    only_return_existing: false,
                    // We should document this as assumed somewhere.
                    terms_of_service_agreed: true,
                },
                &acme.acme_provider,
                None,
            )
            .await?;

            File::create(&acme.acme_account_file)?
                .write_all(serde_json::to_string(&account_creds)?.as_bytes())?;

            account
        }
    };

    let hosts: Vec<String> = dynconfig.read().await.host_map.keys().cloned().collect();
    let mut order = account
        .new_order(&NewOrder {
            identifiers: &hosts
                .iter()
                .map(|h| Identifier::Dns(h.clone()))
                .collect::<Vec<Identifier>>(),
        })
        .await?;

    let mut cur_tokens: Vec<(String, String)> = vec![];
    let mut challenges_to_validate: Vec<String> = vec![];
    for auth in &order.authorizations().await? {
        if auth.status == AuthorizationStatus::Pending {
            match auth
                .challenges
                .iter()
                .find(|ch| ch.r#type == ChallengeType::Http01)
            {
                None => bail!(
                    "ACME provider didn't provide option for HTTP-01 verification for {:#?}",
                    auth.identifier
                ),
                Some(ch) => {
                    cur_tokens.push((
                        ch.token.clone(),
                        order.key_authorization(ch).as_str().to_owned(),
                    ));
                    challenges_to_validate.push(ch.url.clone());
                }
            }
        }
    }

    dynconfig.write().await.acme_challenge_tokens = cur_tokens.into_iter().collect();

    for ch in &challenges_to_validate {
        order.set_challenge_ready(ch).await?;
    }

    loop {
        sleep(Duration::seconds(10).unsigned_abs()).await;
        let state = order.refresh().await?;
        if state.status == OrderStatus::Invalid {
            bail!("Certificate order went to Invalid status");
        }
        if state.status == OrderStatus::Ready {
            break;
        }
        warn!("ACME validation is still not ready 10s after requesting validation. Trying again in 10s.");
    }

    // We are now fully validated, and ready to generate a CSR...
    let mut cert_params: CertificateParams = CertificateParams::new(hosts)?;
    cert_params.distinguished_name = DistinguishedName::new();
    let keypair: KeyPair = KeyPair::generate()?;
    let csr = cert_params.serialize_request(&keypair)?;
    order.finalize(csr.der()).await?;

    let cert = loop {
        sleep(Duration::seconds(1).unsigned_abs()).await;
        match order.certificate().await? {
            None => {
                warn!("ACME certificate still not ready after 10s. Waiting another 10s.");
            }
            Some(cert) => break cert,
        }
    };

    File::create(&tls.certificate_chain_file)?.write_all(cert.as_bytes())?;
    File::create(&tls.private_key_file)?.write_all(keypair.serialize_pem().as_bytes())?;

    rehash_config(config_file, dynconfig).await?;

    Ok(())
}

#[get("/.well-known/acme-challenge/{token}")]
pub async fn acme_challenge(
    path: web::Path<String>,
    config_data: Data<DynamicConfigLock>,
) -> impl Responder {
    let resp = config_data
        .read()
        .await
        .acme_challenge_tokens
        .get(&path.into_inner())
        .cloned();
    match resp {
        None => Err(ErrorNotFound("No such token expected")),
        Some(resp) => Ok(resp),
    }
}

pub async fn start_acme_if_appropriate(
    config_file: &str,
    dynconfig: &DynamicConfigLock,
    config: &ServerConfig,
) {
    let config = config.clone();
    let config_file = config_file.to_owned();
    let dynconfig = dynconfig.clone();
    tokio::spawn(async move {
        if let Some(tls) = &config.tls {
            if let Some(acme) = &tls.acme {
                loop {
                    if let Ok(expiry) = get_current_expiry(&tls.certificate_chain_file) {
                        let check_horizon = OffsetDateTime::now_utc()
                            + Duration::days(acme.renew_if_days_left as i64);
                        if expiry >= check_horizon {
                            let time_left = expiry - check_horizon;
                            info!("Certificate is valid for {} - rechecking then.", time_left);
                            sleep(time_left.unsigned_abs()).await;
                            continue;
                        }
                        info!("Certificate found but due for expiry.");
                    } else {
                        info!("No valid certificate - requesting one.");
                    }

                    if let Err(e) = try_renew_certificate(&config_file, &dynconfig, tls, acme).await
                    {
                        error!(
                            "Renewing certificate failed with error: {}. Trying again in one day.",
                            e
                        );
                        sleep(Duration::days(1).unsigned_abs()).await;
                    }
                }
            }
        }
    });
}
