use actix_files::Files;
use actix_web::{
    self,
    body::BoxBody,
    dev::{
        self, HttpServiceFactory, ResourceDef, Service, ServiceFactory, ServiceRequest,
        ServiceResponse,
    },
    error::ErrorInternalServerError,
    Error,
};
use anyhow::anyhow;
use futures_util::future::LocalBoxFuture;
use log::error;
use rustls::{
    crypto::{aws_lc_rs, CryptoProvider},
    pki_types::CertificateDer,
    server::ResolvesServerCert,
    sign::CertifiedKey,
};
use std::{
    collections::BTreeMap,
    fs::File,
    future::{ready, Ready},
    io::{self, BufReader},
    sync::Arc,
};
use tokio::sync::RwLock;
use wildmatch::WildMatch;

use crate::{load_config_file, validate_config_file, MudConfig, ServerConfig, TlsConfig};

#[derive(Debug)]
pub struct DynamicConfigData {
    pub host_map: BTreeMap<String, MudConfig<WildMatch>>,
    pub certified_key: Option<Arc<CertifiedKey>>,
    pub static_root: String,
    pub acme_challenge_tokens: BTreeMap<String, String>,
}
pub type DynamicConfigLock = Arc<RwLock<DynamicConfigData>>;

#[derive(Debug)]
pub struct DynconfigCertResolver(pub DynamicConfigLock);

impl ResolvesServerCert for DynconfigCertResolver {
    fn resolve(
        &self,
        _client_hello: rustls::server::ClientHello<'_>,
    ) -> Option<Arc<rustls::sign::CertifiedKey>> {
        self.0
            .try_read()
            .ok()
            .and_then(|v| v.certified_key.as_ref().cloned())
    }
}

#[derive(Clone)]
pub struct DynStaticFiles(pub DynamicConfigLock);
impl HttpServiceFactory for DynStaticFiles {
    fn register(self, config: &mut actix_web::dev::AppService) {
        let rdef = if config.is_root() {
            ResourceDef::root_prefix("/")
        } else {
            ResourceDef::prefix("/")
        };

        config.register_service(rdef, None, self, None)
    }
}

impl ServiceFactory<ServiceRequest> for DynStaticFiles {
    type Response = ServiceResponse;
    type Error = Error;
    type Config = ();
    type Service = DynStaticFiles;
    type InitError = ();
    type Future = Ready<Result<Self::Service, Self::InitError>>;

    fn new_service(&self, _: ()) -> Self::Future {
        ready(Ok(self.clone()))
    }
}

impl Service<ServiceRequest> for DynStaticFiles {
    type Response = ServiceResponse<BoxBody>;
    type Error = Error;
    type Future = LocalBoxFuture<'static, Result<Self::Response, Self::Error>>;

    dev::always_ready!();
    fn call(&self, req: ServiceRequest) -> Self::Future {
        let dynconfig = self.0.clone();
        Box::pin(async move {
            let files =
                Files::new("/", &dynconfig.read().await.static_root).index_file("index.html");
            files
                .new_service(())
                .await
                .map_err(|_| ErrorInternalServerError("Internal service error"))?
                .call(req)
                .await
        })
    }
}

fn load_certified_key(tls: &TlsConfig) -> anyhow::Result<Arc<CertifiedKey>> {
    let certs: Vec<CertificateDer> = rustls_pemfile::certs(&mut BufReader::new(&mut File::open(
        &tls.certificate_chain_file,
    )?))
    .collect::<Result<Vec<CertificateDer>, io::Error>>()?;
    let key =
        rustls_pemfile::private_key(&mut BufReader::new(&mut File::open(&tls.private_key_file)?))?
            .ok_or_else(|| anyhow!("No private key found in private key file"))?;
    let key = aws_lc_rs::default_provider()
        .key_provider
        .load_private_key(key)?;
    CryptoProvider::get_default();
    Ok(CertifiedKey::new(certs, key).into())
}

pub fn make_dynamic_config_data(config: &ServerConfig) -> anyhow::Result<DynamicConfigData> {
    Ok(DynamicConfigData {
        host_map: config
            .muds
            .iter()
            .map(|c| {
                (
                    c.hostname.clone(),
                    MudConfig::<WildMatch> {
                        hostname: c.hostname.clone(),
                        upstream_mud: c.upstream_mud.clone(),
                        banner_to_mud: c.banner_to_mud.clone(),
                        startup_script_file: c.startup_script_file.clone(),
                        allowed_origins: c
                            .allowed_origins
                            .iter()
                            .map(|ao| WildMatch::new(ao))
                            .collect(),
                    },
                )
            })
            .collect::<BTreeMap<String, MudConfig<WildMatch>>>(),
        certified_key: match &config.tls {
            None => None,
            Some(tls) => match load_certified_key(tls) {
                Ok(ck) => Some(ck),
                Err(e) => {
                    error!(
                        "Couldn't load TLS certs: {}. Will continue in case updated data \
                            is coming after rehash, but the server won't work for now.",
                        e
                    );
                    None
                }
            },
        },
        static_root: config.static_root.clone(),
        acme_challenge_tokens: BTreeMap::new(),
    })
}

pub async fn rehash_config(
    config_file: &str,
    dynconfig_lock: &DynamicConfigLock,
) -> anyhow::Result<()> {
    let config: ServerConfig = load_config_file(config_file)?;
    validate_config_file(&config);
    let dynconfig: DynamicConfigData = make_dynamic_config_data(&config)?;

    *dynconfig_lock.write().await = dynconfig;

    Ok(())
}
