use acme::{acme_challenge, start_acme_if_appropriate};
use actix_web::{middleware::Logger, web::Data, App, HttpServer};
use dynamic_config::{
    make_dynamic_config_data, rehash_config, DynStaticFiles, DynamicConfigLock,
    DynconfigCertResolver,
};
use futures_util::FutureExt;
use log::{info, warn};
use serde::Deserialize;
use std::{env, fs, sync::Arc};
use tokio::{
    select,
    signal::unix::{signal, SignalKind},
    spawn,
    sync::{oneshot, RwLock},
};
use ws_svc::ws;

mod acme;
mod dynamic_config;
mod ws_svc;

#[derive(Deserialize, Clone)]
#[serde(default)]
pub struct TlsConfig {
    pub private_key_file: String,
    pub certificate_chain_file: String,
    pub acme: Option<AcmeConfig>,
}

impl Default for TlsConfig {
    fn default() -> Self {
        Self {
            private_key_file: "/etc/privkey.pem".to_owned(),
            certificate_chain_file: "/etc/fullchain.pem".to_owned(),
            acme: None,
        }
    }
}

#[derive(Deserialize, Clone)]
#[serde(default)]
pub struct AcmeConfig {
    pub contact_email: String,
    pub acme_account_file: String,
    pub acme_provider: String,
    pub renew_if_days_left: u8,
}

impl Default for AcmeConfig {
    fn default() -> Self {
        Self {
            contact_email: "".to_owned(),
            acme_account_file: "/etc/acme_account.txt".to_owned(),
            acme_provider: "https://acme-v02.api.letsencrypt.org/directory".to_owned(),
            renew_if_days_left: 30,
        }
    }
}

#[derive(Deserialize, Clone)]
#[serde(default)]
pub struct ServerConfig {
    pub listen_port: u16,
    pub bind_address: String,
    pub tls: Option<TlsConfig>,
    pub static_root: String,
    pub muds: Vec<MudConfig<String>>,
}

impl Default for ServerConfig {
    fn default() -> Self {
        Self {
            listen_port: 443,
            bind_address: "::".to_owned(),
            tls: None,
            static_root: "/frontendstatic".to_owned(),
            muds: vec![],
        }
    }
}

#[derive(Deserialize, Clone, Debug)]
pub struct MudConfig<WildType> {
    pub hostname: String,
    pub upstream_mud: String,
    pub banner_to_mud: String,
    pub startup_script_file: String,
    pub allowed_origins: Vec<WildType>,
}

// We load this on each connection so it can change.
fn fetch_startup_script(filename: &str) -> anyhow::Result<String> {
    Ok(fs::read_to_string(filename)?)
}

fn load_config_file(filename: &str) -> anyhow::Result<ServerConfig> {
    info!("Reading config file {}", filename);
    Ok(toml::from_str(&fs::read_to_string(filename)?)?)
}

fn validate_config_file(config: &ServerConfig) {
    match fs::metadata(&config.static_root).map(|meta| meta.is_dir()) {
        Err(e) => warn!(
            "Error checking static_root {} is valid: {}",
            &config.static_root, e
        ),
        Ok(false) => warn!("static_root {} is not a directory", &config.static_root),
        Ok(true) => {}
    }
    for mud in &config.muds {
        match fetch_startup_script(&mud.startup_script_file) {
            Err(e) => warn!(
                "Error reading startup script file {} for MUD {}: {}",
                mud.startup_script_file, mud.hostname, e
            ),
            Ok(_) => {}
        }
    }
}

#[tokio::main]
async fn main() -> anyhow::Result<()> {
    env_logger::builder()
        .filter_level(log::LevelFilter::Info)
        .parse_default_env()
        .init();
    let _ = rustls::crypto::aws_lc_rs::default_provider().install_default();
    let args: Vec<String> = env::args().collect();
    let config_file = match args.get(1) {
        None => "/etc/worldwideportal_server.toml".to_owned(),
        Some(cfg) => cfg.clone(),
    };
    if config_file == "-h" || config_file == "--help" {
        println!("Usage: worldwideportal-server /path/to/config/file.toml");
        return Ok(());
    }

    let config: ServerConfig = load_config_file(&config_file)?;
    validate_config_file(&config);
    let dynconfig: DynamicConfigLock = RwLock::new(make_dynamic_config_data(&config)?).into();

    start_acme_if_appropriate(&config_file, &dynconfig, &config).await;

    // This is a double Arc, but actix doesn't provide an alternative.
    let data = Data::new(dynconfig.clone());

    let server_data = data.clone();
    let static_file_svc = DynStaticFiles(dynconfig.clone());
    let server = HttpServer::new(move || {
        let logger = Logger::default();
        App::new()
            .wrap(logger)
            .app_data(server_data.clone())
            .default_service(static_file_svc.clone())
            .service(ws)
            .service(acme_challenge)
    });

    let server = match &config.tls {
        None => server.bind((config.bind_address.clone(), config.listen_port))?,
        Some(tls) => {
            let tls_config = rustls::server::ServerConfig::builder()
                .with_no_client_auth()
                .with_cert_resolver(Arc::new(DynconfigCertResolver(dynconfig.clone())));

            let server = if tls.acme.is_some() {
                // We always bind port 80 if they are using ACME - it is essential for
                // the http-01 challenge to work (and must be that port).
                server.bind((config.bind_address.clone(), 80))?
            } else {
                server
            };

            server.bind_rustls_0_23(
                (config.bind_address.clone(), config.listen_port),
                tls_config,
            )?
        }
    };

    let mut sighup = signal(SignalKind::hangup())?;
    let (exit_tx, exit_rx) = oneshot::channel::<()>();
    let exit_rx = exit_rx.shared();
    spawn(async move {
        loop {
            select! {
                _ = sighup.recv() => {
                    let _ = rehash_config(&config_file, &dynconfig).await;
                }
                _ = exit_rx.clone() => {
                    return;
                }
            }
        }
    });

    server.run().await?;
    let _ = exit_tx.send(());
    Ok(())
}
