use crate::{dynamic_config::DynamicConfigLock, fetch_startup_script, MudConfig};
use actix_web::{
    error::{ErrorBadRequest, ErrorForbidden, ErrorNotFound, InternalError},
    get,
    http::{
        header::{HOST, ORIGIN},
        StatusCode,
    },
    rt::{self, net::TcpStream},
    web::{self, Data},
    Error, HttpRequest, HttpResponse, Responder,
};
use actix_ws::{AggregatedMessage, CloseCode, Closed};
use futures_util::StreamExt;
use serde_json::json;
use tokio::io::{AsyncReadExt, AsyncWriteExt};
use wildmatch::WildMatch;

async fn with_mudconfig(
    req: &HttpRequest,
    config_data: Data<DynamicConfigLock>,
) -> Result<MudConfig<WildMatch>, Error> {
    match req.headers().get(&HOST) {
        None => Err(ErrorBadRequest("Missing Host header"))?,
        Some(host) => match config_data
            .read()
            .await
            .host_map
            .get(host.to_str().unwrap_or("invalid"))
        {
            None => Err(ErrorNotFound(
                "No MUD matching your request is currently enabled",
            ))?,
            Some(v) => Ok(v.clone()),
        },
    }
}

#[get("/ws")]
pub async fn ws(
    config_data: Data<DynamicConfigLock>,
    req: HttpRequest,
    body: web::Payload,
) -> impl Responder {
    let mud = with_mudconfig(&req, config_data).await?;
    match req.headers().get(&ORIGIN) {
        None => Err(ErrorForbidden("Missing origin"))?,
        Some(origin) => {
            if !mud
                .allowed_origins
                .iter()
                .any(|o| o.matches(origin.to_str().unwrap_or("invalid")))
            {
                Err(ErrorForbidden("Disallowed origin"))?;
            }
        }
    }

    let (response, mut session, stream) = actix_ws::handle(&req, body)?;
    let mut stream = stream.aggregate_continuations().max_continuation_size(1024);

    let mut tcp_stream: TcpStream = TcpStream::connect(&mud.upstream_mud).await?;
    let subst_banner = mud
        .banner_to_mud
        .replace(
            "%i",
            &req.peer_addr()
                .map(|a| a.ip().to_string())
                .unwrap_or_else(|| "unknown".to_owned()),
        )
        .replace("%n", "\r\n");
    tcp_stream.write_all(subst_banner.as_bytes()).await?;

    let script = fetch_startup_script(&mud.startup_script_file)
        .map_err(|e| InternalError::new(e, StatusCode::INTERNAL_SERVER_ERROR))?;

    rt::spawn(async move {
        if session
            .text(
                json!({
                    "RunLua": script
                })
                .to_string(),
            )
            .await
            .is_err()
        {
            let _ = session.close(Some(CloseCode::Normal.into())).await;
            return;
        }
        let mut readbuf: [u8; 1024] = [0; 1024];
        loop {
            tokio::select! {
                ws_msg = stream.next() => {
                    match ws_msg {
                        None => break,
                        Some(Err(_e)) => break,
                        Some(Ok(AggregatedMessage::Binary(bin))) => {
                            if tcp_stream.write_all(&bin).await.is_err() {
                                break
                            }
                        }

                        Some(Ok(AggregatedMessage::Ping(msg))) => {
                            if let Err(Closed) = session.pong(&msg).await {
                                break
                            }
                        }
                        Some(Ok(_)) => {},
                    }
                },
                tcp_data_len = tcp_stream.read(&mut readbuf) => {
                    match tcp_data_len {
                        Err(_e) => break,
                        Ok(0) => break,
                        Ok(n) =>
                            if let Err(Closed) = session.binary(readbuf[0..n].to_vec()).await {
                                break;
                            }
                    }

                }
            }
        }
        let _ = session.close(Some(CloseCode::Normal.into())).await;
    });

    Ok::<HttpResponse, Error>(response)
}
